"""Releases are bundles of objects of type :class:`batzenca.database.keys.Key` for a given
:class:`batzenca.database.mailinglists.MailingList`. "Making releases" is what this library is for.

.. module:: releases

.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

"""
import os
import datetime

import warnings
import codecs

import sqlalchemy
from sqlalchemy import Column, Integer, String, Date, Boolean, ForeignKey
from sqlalchemy.orm import relationship, backref, Session
from sqlalchemy.ext.associationproxy import association_proxy

from base import Base, EntryNotFound
from peers import Peer
from keys import Key

class ReleaseKeyAssociation(Base):
    __tablename__ = 'releasekeyassociations'

    left_id          = Column(Integer, ForeignKey('keys.id'),     primary_key=True)
    right_id         = Column(Integer, ForeignKey('releases.id'), primary_key=True)
    policy_exception = Column(Boolean)
    is_active        = Column(Boolean)

    key              = relationship("Key", backref=backref("release_associations", cascade="all, delete-orphan") )
    release          = relationship("Release", backref=backref("key_associations", cascade="all, delete-orphan") )

    def __init__(self, key, active=True, policy_exception=False):
        self.key = key
        self.is_active = active
        self.policy_exception = policy_exception

class Release(Base):
    """Releases are bundles of objects of type :class:`batzenca.database.keys.Key`.

    Releases contain active and inactive keys. The former are keys users are expected to use. The
    latter inform the user about invalidated keys for example by revoked signatures.
    
    :param batzenca.database.mailinglists.MailingList mailinglist: the mailinglist for which this release
        is intended
    :param date: the date of this release
    :param iterable active_keys: keys distributed in this release that user ought to use
    :param iterable inactive_keys: keys which are not active in this release, yet should be
        distributed. For example, this could include keys with revocation signatures which are
        distributed to inform users about this revocation.
    :param batzenca.database.policies.Policy policy: the policy against which keys in this release
        should be checked

    """
    __tablename__ = 'releases'

    id             = Column(Integer, primary_key=True)
    mailinglist_id = Column(Integer, ForeignKey('mailinglists.id'))
    mailinglist    = relationship("MailingList", backref=backref("releases", order_by="Release.date", cascade="all, delete-orphan"))
    date           = Column(Date)
    published      = Column(Boolean)

    policy_id      = Column(Integer, ForeignKey('policies.id'))
    policy         = relationship("Policy")

    keys           = association_proxy('key_associations', 'key')

    def __init__(self, mailinglist, date, active_keys, inactive_keys=None, policy=None):
        self.mailinglist = mailinglist

        if date is None:
            date = datetime.date.today()
        self.date = date

        if policy is not None:
            self.policy = policy
        else:
            self.policy = mailinglist.policy

        for key in active_keys:
            self.key_associations.append(ReleaseKeyAssociation(key=key))

        for key in inactive_keys:
            self.key_associations.append(ReleaseKeyAssociation(key=key, active=False))

        self.published = False

    @classmethod
    def from_mailinglist_and_date(cls, mailinglist, date):
        """Return the release on ``mailinglist`` for ``date`` from the database.  If more than one
        element is found the "first" element is returned, where "first" has no particular meaning
        and is implementation specific.  In this case a warning is issued.

        
        :param batzenca.database.mailinglists.MailingList mailinglist: the mailinglist on which the
            target release was released
        :param date: the date on which the target release was released
        
        :raises batzenca.database.base.EntryNotFound: when no entry is found

        .. note::

           The returned object was aquired from the master session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(cls).filter(cls.mailinglist_id == mailinglist.id, cls.date == date)

        if res.count() == 0:
            raise EntryNotFound("No release for mailinglist '%s' with date '%s' in database."%(mailinglist, date))
        else:
            if res.count() > 1:
                warnings.warn("More than one release for mailinglist '%s' with date '%s' in database, picking first one"%(mailinglist, date))
            return res.first()

    def inherit(self, date=None, policy=None, deactivate_invalid=True, delete_old_inactive_keys=5):
        """Construct a new release by inheritance from this release. Inheritance means that active
        and inactive keys are carried forward.
        
        :param date: the date of this release. If ``None`` today's date is used
        :param boolean deactivate_invalid: deactivate keys which are no longer valid, e.g. because
          they are expired.
        :param boolean delete_old_inactive_keys: delete inactive keys which have been around for a
          while, this parameter is passed to
          :func:`batzenca.database.releases.Release.delete_old_inactive_keys` as
          ``releasecount``.

        """
        active_keys   = list(self.active_keys)
        inactive_keys = list(self.inactive_keys)

        if policy is None:
            policy = self.policy
            
        release = Release(mailinglist=self.mailinglist, 
                          date=date, 
                          active_keys = active_keys, 
                          inactive_keys = inactive_keys, 
                          policy=policy)

        if deactivate_invalid:
            release.deactivate_invalid() 
        if delete_old_inactive_keys:
            release.delete_old_inactive_keys(delete_old_inactive_keys)
            
        for key in self.active_keys:
            if self.has_exception(key):
                release.add_exception(key)

        return release

    def verify(self, ignore_exceptions=False):
        """Check if all active keys in this release pass the policy check.

        :param boolean ignore_exceptions: keys may have a policy exception which means that they
           pass this test even though they do violate the policy. By default active keys with an
           existing policy exception are ignored. If ``True`` these keys are checked as well.

        """
        for assoc in self.key_associations:
            if assoc.is_active and (ignore_exceptions or not assoc.policy_exception):
                self.policy.check(assoc.key)

    def __repr__(self):
        s = "<Release: %s, %s (%s), %s (%s + %s) keys>"%(self.id, self.date, self.mailinglist, len(self.key_associations), len(self.active_keys), len(self.inactive_keys))
        return unicode(s).encode('utf-8')
                
    def __str__(self):
        from batzenca.database.policies import PolicyViolation
        inact_no_sig = 0
        inact_expired = 0
        policy = self.policy
        for key in self.keys:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", PolicyViolation)
                if policy.check_ca_signature(key) == False:
                    inact_no_sig += 1
                    continue

            if key.expires and key.expires < self.date:
                inact_expired += 1
                continue

        return "date: %10s, list: %10s, policy date: %10s, active keys: %3d, inactive keys: %2d (expired: %2d, not signed: %2d), total keys: %3d"%(self.date, self.mailinglist.name, self.policy.implementation_date,
                                                                                                                                                   len(self.active_keys), len(self.inactive_keys), inact_expired, inact_no_sig, len(self.keys))

    @property
    def ascii_keys(self):
        """All active and inactive keys in this release as OpenPGP ASCII text"""
        from batzenca.session import session
        return session.gnupg.keys_export([key.kid for key in self.keys])

    def diff(self, other=None):
        """Compare this release with ``other``.

        :param batzenca.database.releases.Release other: the release to compare
            against, if ``None`` then ``self.prev`` is chosen
        
        :return: this function returns five tuples:

        - ``keys_in`` - keys that are active in this release but are not active in ``other``
        - ``keys_out`` - all keys that are either active or inactive in ``other`` but are not active
          in this release.
        - ``peers_joined`` - peers active in this release but not in ``other``
        - ``peers_changed`` - peers that have different active keys in ``other`` and this release
        - ``peers_left`` - peers that are active in ``other`` but in this release

        """
        if other is None:
            other = self.prev
        
        keys_prev = set(other.active_keys + self.inactive_keys)        
        keys_curr = set(self.active_keys) # keys that are in this release

        # keys that used to be in but are not any more
        keys_out = keys_prev.difference(keys_curr)
        # keys that are new
        keys_in  = keys_curr.difference(other.active_keys)
        
        peers_prev = set([Peer.from_key(key) for key in keys_prev])
        peers_curr = set([Peer.from_key(key) for key in keys_curr])
        peers_in   = set([Peer.from_key(key) for key in keys_in  ])
        peers_out  = set([Peer.from_key(key) for key in keys_out ])

        peers_joined  = peers_curr.difference(peers_prev)
        peers_changed = peers_in.intersection(peers_out)
        peers_left    = peers_prev.difference(peers_curr)


        return keys_in, keys_out, peers_joined, peers_changed, peers_left

    @property
    def peers(self):
        """All active peers in this release"""
        return tuple(Peer.from_key(key) for key in sorted(self.active_keys, key=lambda x:
                                                          x.name.lower()))

    @staticmethod
    def _format_entry(i, key):
        return (u"  %3d. %s"%(i, key), u"       %s"%key.peer)

    @property
    def active_keys(self):
        """All active keys in this release."""
        if self.id is None:
            return [assoc for assoc in self.key_associations if assoc.is_active]
        from batzenca.session import session
        return session.db_session.query(Key).join(ReleaseKeyAssociation).filter(ReleaseKeyAssociation.right_id == self.id, ReleaseKeyAssociation.is_active == True).all()

    @property
    def inactive_keys(self):
        """All inactive keys in this release."""
        if self.id is None:
            return [assoc.key for assoc in self.key_associations if not assoc.is_active]
        from batzenca.session import session
        return session.db_session.query(Key).join(ReleaseKeyAssociation).filter(ReleaseKeyAssociation.right_id == self.id, ReleaseKeyAssociation.is_active == False).all()

    def deactivate_invalid(self):
        """Deactivate those keys which evaluate to false and those keys which are not signed by the
        CA.

        A key evaluates to false if it is expired. Keys are considered unsigned if the CA signature
        is revoked.

        """
        if self.published:
            raise ValueError("Release '%s' is already published and should not be modified."%self)

        for assoc in self.key_associations:
            if assoc.is_active:
                if not bool(assoc.key):
                    assoc.is_active = False
                elif not assoc.key.is_signed_by(self.policy.ca):
                    assoc.is_active = False

    def delete_old_inactive_keys(self, releasecount=5):
        """Remove those inactive keys which have been inactive for a while.

        :param boolean releasecount: the number of releases for which a key must have been inactive
            to be removed.

        """
        if self.published:
            raise ValueError("Release '%s' is already published and should not be modified."%self)

        old_release = self
        for i in range(releasecount):
            if old_release.prev:
                old_release = old_release.prev
            else:
                return

        delete_keys = []
        for key in self.inactive_keys:
            if key not in old_release.active_keys:
                delete_keys.append(key)
            elif key.expires and key.expires < self.date:
                delete_keys.append(key)
        for key in delete_keys:
            assoc = self._get_assoc(key)
            self.key_associations.remove(assoc)
            from batzenca.session import session
            session.db_session.delete(assoc)
            
                
    def _get_assoc(self, key):
        if key.id is None or self.id is None:
            for assoc in self.key_associations:
                if assoc.key is key and assoc.release is self:
                    return assoc
            raise ValueError("Key '%s' is not in release '%s'"%(key, self))

        from batzenca.session import session
        res = session.db_session.query(ReleaseKeyAssociation).filter(ReleaseKeyAssociation.left_id == key.id, ReleaseKeyAssociation.right_id == self.id)
        if res.count() > 1:
            raise RuntimeError("The key '%s' is associated with the release '%' more than once; the database is in an inconsistent state."%(key, self))
        if res.count() == 0:
            raise ValueError("Key '%s' is not in release '%s'"%(key, self))
        return res.first()

    def add_exception(self, key):
        """Add a policy exception for the provided key.

        :param batzenca.database.keys.Key key: the key for which to add the exception

        """
        if self.published:
            raise ValueError("Release '%s' is already published and should not be modified."%self)

        assoc = self._get_assoc(key)
        assoc.policy_exception = True

    def has_exception(self, key):
        """Return ``True`` if the provided key has a policy exception.

        :param batzenca.database.keys.Key key: the key to check

        """
        assoc = self._get_assoc(key)
        return assoc.policy_exception

    def is_active(self, key):
        """Return ``True`` if the provided key is active in this release,

        :param batzenca.database.keys.Key key: the key to check

        """
        assoc = self._get_assoc(Key)
        return assoc.is_active

    def update_key_from_peer(self, peer):
        if not peer in self:
            raise ValueError("Peer '%s' is not in '%s'"%(peer, self))
        if peer.key in self:
            raise ValueError("Key '%s' of peer '%s' is already in release '%s'"%(peer.key, peer, self))

        from batzenca.session import session
        res = session.db_session.query(ReleaseKeyAssociation).join(Key).join(Peer).filter(Key.peer_id == peer.id, ReleaseKeyAssociation.left_id == Key.id, ReleaseKeyAssociation.right_id == self.id, ReleaseKeyAssociation.is_active == True)

        for assoc in res.all():
            if not bool(assoc.key):
                assoc.is_active = False
            elif not assoc.key.is_signed_by(self.policy.ca):
                assoc.is_active = False
            else:
                raise ValueError("Key '%s' of peer '%s' has a valid signature by CA '%s' mandated in '%s'"%(assoc.key, peer, self.policy.ca, self.policy))

        self.add_key(peer.key, active=True, check=True)

    def add_key(self, key, active=True, check=True):
        if self.published:
            raise ValueError("Release '%s' is already published and should not be modified."%self)

        if key.peer is None:
            raise ValueError("Key '%s' has no peer associated"%key)
        else:
            if active and key.peer in self:
                raise ValueError("Peer '%s' associated with Key '%s' already has an active key in this release"%(key.peer, key))
            
        if check and active:
            self.policy.check(key)
                        
        self.key_associations.append(ReleaseKeyAssociation(key=key, active=active))

    def __contains__(self, obj):
        from batzenca.session import session

        if self.id is None:
            raise RuntimeError("The object '%s' was not committed to the database yet, we cannot issue queries involving its id yet."%self)

        try:
            if obj.id is None:
                raise RuntimeError("The object '%s' was not committed to the database yet, we cannot issue queries involving its id yet."%obj)
        except AttributeError:
            raise TypeError("Cannot handle objects of type '%s'"%type(obj))

        if isinstance(obj, Key):
            res = session.db_session.query(Key).join(ReleaseKeyAssociation).filter(ReleaseKeyAssociation.left_id == obj.id, ReleaseKeyAssociation.right_id == self.id, ReleaseKeyAssociation.is_active == True)
            if res.count() == 0:
                return False
            elif res.count() == 1:
                return True
            else:
                raise RuntimeError("The key '%s' is associated with the release '%' more than once; the database is in an inconsistent state."%(obj, self))
            
        elif isinstance(obj, Peer):
            res = session.db_session.query(Peer).join(Key).join(ReleaseKeyAssociation).filter(Key.peer_id == obj.id, ReleaseKeyAssociation.left_id == Key.id, ReleaseKeyAssociation.right_id == self.id, ReleaseKeyAssociation.is_active == True)
            if res.count() == 0:
                return False
            elif res.count() == 1:
                return True
            else:
                raise RuntimeError("The peer '%s' is associated with the release '%' more than once; the database is in an inconsistent state."%(obj, self))
        else:
            raise TypeError("Cannot handle objects of type '%s'"%type(obj))

    @property
    def prev(self):
        """
        The previous release.
        """
        idx = self.mailinglist.releases.index(self)
        if idx > 0:
            return self.mailinglist.releases[idx-1]
        else:
            return None

    @property
    def yaml(self):
        """YAML representation of this release.

        This contains the name of the database, the date of this release, the key id of the CA, all
        key ids of active and passive keys.

        """
        s = []
        s.append( "mailinglist: %s"%self.mailinglist.email )
        s.append( "date:        %04d-%02d-%02d"%(self.date.year, self.date.month, self.date.day) )
        s.append( "ca:          %s"%self.policy.ca.kid )
        
        s.append( "active keys:" )
        for key in self.active_keys:
            s.append( "  - %s"%key.kid )
        s.append( "" )
        s.append( "inactive keys:" )
        for key in self.inactive_keys:
            s.append( "  - %s"%key.kid )
        s.append( "" )
        return "\n".join(s)

    def expiring_keys(self, days=30):
        return tuple(key for key in self.active_keys if key.expires and key.expires < self.date + datetime.timedelta(days=days))
        
    def __call__(self, previous=None, check=True, still_alive=False):
        """Return tuple representing this release as a (message, keys) pair.

        :param batzenca.database.releases.Release previous: the previous release, we call
            :func:`batzenca.database.releases.Release.diff` on it.  if ``None`` then ``self.prev``
            is used.
        :param boolean check: if ``True`` then :func:`batzenca.database.releases.Release.verify` is
            run.

        :param boolean still_alive: if ``True`` then the field ``{dead_man_switch}`` in
            :attr:`batzenca.database.mailinglists.MailingList.key_update_msg` is replaced by
            :attr:`batzenca.database.mailinglists.MailingList.dead_man_switch_msg`.

        :return: a tuple containing two strings. The first one is
            :attr:`batzenca.database.mailingists.MailingList.key_update_msg` with the output of
            :class:`batzenca.database.releases.Release.diff` used to fill in its fields. The second
            component is ``self``'s :attr:`batzenca.database.releases.Release.ascii_keys`.

        """
        if check:
            self.verify()

        sorted_keys = lambda keys: sorted(keys, key=lambda x: x.name.lower())
            
        keys = []
        for i,key in enumerate(sorted_keys(self.active_keys)):
            keys.extend(Release._format_entry(i, key))
        keys = "\n".join(keys)

        if previous is None:
            previous = self.prev

        if previous:
            keys_in, keys_out, peers_joined, peers_changed, peers_left = self.diff(previous)

            keys_in  = "\n".join(sum([self._format_entry(i, key) for i,key in enumerate(sorted_keys(keys_in)) ], tuple()))
            keys_out = "\n".join(sum([self._format_entry(i, key) for i,key in enumerate(sorted_keys(keys_out))], tuple()))

            peers_joined  = ", ".join(peer.name for peer in peers_joined)
            peers_changed = ", ".join(peer.name for peer in peers_changed)
            peers_left    = ", ".join(peer.name for peer in peers_left)
        else:
            keys_in, keys_out, peers_joined, peers_changed, peers_left = "","","","",""
            
        msg = self.mailinglist.key_update_msg.format(mailinglist=self.mailinglist.name, keys=keys,
                                                     keys_in       = keys_in,
                                                     keys_out      = keys_out,
                                                     peers_in      = peers_joined,
                                                     peers_changed = peers_changed,
                                                     peers_out     = peers_left,
                                                     dead_man_switch = self.mailinglist.dead_man_switch_msg if still_alive else "",
                                                     ca=self.policy.ca.name,
                                                     ca_email=self.policy.ca.email)

        return msg, self.ascii_keys

    def release_message(self, previous=None, check=True, debug=False, attachments=None):
        """
        """
        ca = self.policy.ca
        mailinglist = self.mailinglist
        date_str = "%04d%02d%02d"%(self.date.year, self.date.month, self.date.day)

        body_, attachment_ = self(previous=previous, check=check)

        from email import encoders
        from email.mime.base import MIMEBase
        from email.mime.multipart import MIMEMultipart
        from email.mime.text import MIMEText
        from batzenca.pgpmime import PGPMIME
        
        payload = MIMEMultipart()
        payload.attach(MIMEText(body_.encode('utf-8'),  _charset='utf-8'))

        attachment = MIMEBase('application', 'pgp-keys')
        attachment.set_payload(attachment_)
        encoders.encode_base64(attachment)
        attachment.add_header('Content-Disposition', 'attachment', filename="%s_%s.asc"%(mailinglist.name, date_str))
        payload.attach(attachment)

        if attachments:
            for attachment in attachments:
                payload.attach(attachment)

        msg = PGPMIME(payload, self.active_keys, ca)
    
        # we are being a bit paranoid and check that we didn't fuck up encryption or something
        for key in self.active_keys:
            assert(key.kid not in msg.as_string())

        to = mailinglist.email if not debug else ca.email
            
        msg['To']      = to
        msg['From']    = ca.email
        msg['Subject'] = "KeyUpdate {date} [{mailinglist}]".format(date=date_str, mailinglist=mailinglist.name)

        return msg
        
    def welcome_messages(self, tolerance=180, debug=False):
        mailinglist = self.mailinglist
        ca = self.policy.ca
        today = datetime.date.today()
        tolerance = today - datetime.timedelta(days=tolerance)

        from email.mime.text import MIMEText
        from batzenca.pgpmime import PGPMIME


        M = []
        for peer in self.peers:
            was_recently_active = False
            for key in peer.keys:
                if any(rel.date > tolerance for rel in key.releases if rel.mailinglist == self.mailinglist and rel != self):
                    was_recently_active = True
                    break
            if not was_recently_active:
                body    = self.mailinglist.new_member_msg.format(peer=peer.name,
                                                                 mailinglist=mailinglist.name,
                                                                 mailinglist_email=mailinglist.email,
                                                                 ca=ca.name,
                                                                 ca_email=ca.email)
                payload = MIMEText(body.encode('utf-8'),  _charset='utf-8')
                msg = PGPMIME(payload, [peer.key, ca], ca)
                to = peer.email if not debug else ca.email
                msg['To']      = to
                msg['From']    = ca.email
                msg['Subject'] = "welcome to [{mailinglist}]".format(mailinglist=mailinglist.name)
                M.append( msg )
        return tuple(M)

    def key_expiry_messages(self, days=30, debug=False):
        mailinglist = self.mailinglist
        ca = self.policy.ca

        from email.mime.text import MIMEText
        from batzenca.pgpmime import PGPMIME

        M = []
        for key in self.expiring_keys(days=days):
            body    = self.mailinglist.key_expiry_warning_msg.format(peer=key.peer.name,
                                                                     keyid=key.kid,
                                                                     expiry_date=key.expires,
                                                                     mailinglist = mailinglist.name,
                                                                     mailinglist_email = mailinglist.email,
                                                                     ca_email = ca.email)
            payload = MIMEText(body.encode('utf-8'),  _charset='utf-8')
            msg = PGPMIME(payload, [key, ca], ca)
            to = key.email if not debug else ca.email
            msg['To']      = to
            msg['From']    = ca.email
            msg['Subject'] = "key expiry warning".format(mailinglist=mailinglist.name)
            M.append( msg )
        return tuple(M)


    def dump(self, filename=None):
        """Write this release to to filename.yaml and filename.asc where the
        former receives :attr:`batzenca.database.releases.Release.yaml` and the
        latter receives :attr:`batzenca.database.releases.Release.ascii_keys`.

        :param str filename: a string containing a full path and basename.

        """
        from batzenca import session
        
        if filename is None:
            filename = os.path.join(session.release_dump_path,
                                    "%s-%04d%02d%02d"%(self.mailinglist.name,
                                                       self.date.year, self.date.month, self.date.day))
        codecs.open(filename+".yaml", encoding="utf8", mode="w").write( self.yaml )
        open(filename+".asc", "w").write( self(previous=None, check=False)[1] )

        
    def send(self, smtpserver, previous=None, check=True, debug=False, attachments=None,
             new_peer_tolerance_days=180, key_expiry_warning_days=30):
        """Publish this release.

        This entails (if ``debug==False``):

        1. updating the release date of this release

        2. a call to :func:`batzenca.database.releases.Release.deactivate_invalid`

        3. sending an e-mail to new members who have not been on this list for ``new_peer_tolerance_days`` days.

        4. sending a key update message to the list

        5. sending a key expiry message to keys that expire within ``key_expiry_warning_days`` days
       
        6. a call to :func:`batzenca.database.releases.Release.dump`

        7. setting this release status to published.

        If ``debug == True`` then e-mails are sent to the CA's e-mail address instead of the list and/or
        peers. Furthermore, neither the date nor the published status of this release is affected in this case.

        :param smtpserver:
        :param batzenca.database.releases.Release previous: the previous release, we call
            :func:`batzenca.database.releases.Release.diff` on it.  if ``None`` then ``self.prev``
            is used.
        :param boolean check: if ``True`` then :func:`batzenca.database.releases.Release.verify` is
            run.
        :param boolean debug:
        :param iterable attachments:
        :param int new_peer_tolerance_days:
        :param int key_expiry_warning_days:

        .. warning:

            Calling this function may modify this release. Firstly, this function calls
            :func:`batzenca.database.releases.Release.deactivate_invalid`. Secondly, if ``debug`` is ``False``,
            :attr:`batzenca.database.releases.Release.date` is set to today's date and
            :attr:`batzenca.database.releases.Release.published` is set to ``True``.

        """
        if self.published:
            raise ValueError("Release '%s' is already published"%self)

        # 1. updating the release date of this release
            
        if not debug:
            self.date = datetime.date.today()

        # 2. a call to :func:`batzenca.database.releases.Release.deactivate_invalid`

        self.deactivate_invalid()

        # 3. sending an e-mail to new members who have not been on this list for ``new_peer_tolerance_days`` days.

        if new_peer_tolerance_days and self.mailinglist.new_member_msg:
            messages = self.welcome_messages(tolerance=new_peer_tolerance_days, debug=debug)
            for msg in messages:
                if debug:
                    smtpserver.sendmail(self.policy.ca.email, (msg['To'],self.policy.ca.email), msg.as_string())
                else: # we send a copy to self
                    smtpserver.sendmail(self.policy.ca.email, (self.policy.ca.email, ), msg.as_string())

        # 4. sending a key update message to the list

        msg = self.release_message(previous=previous, check=check, debug=debug, attachments=attachments)
        smtpserver.sendmail(self.policy.ca.email, (msg['To'],), msg.as_string())
       
        # 5. sending a key expiry message to keys that expire within ``key_expiry_warning_days`` days
                    
        if key_expiry_warning_days and self.mailinglist.key_expiry_warning_msg:
            messages = self.key_expiry_messages(days=key_expiry_warning_days, debug=debug)
            for msg in messages:
                if debug:
                    smtpserver.sendmail(self.policy.ca.email, (self.policy.ca.email,), msg.as_string())
                else:
                    # we send a copy to self
                    smtpserver.sendmail(self.policy.ca.email, (msg['To'],self.policy.ca.email), msg.as_string())

        # 6. a call to :func:`batzenca.database.releases.Release.dump`

        if not debug:
            self.dump()

        # 7. setting this release status to published.

        if not debug:
            self.published = True



from database.base import Base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import os
import git


class Session:

    _gitignore = """.gitignore
gnupg/pubring.gpg~
gnupg/random_seed
releases
"""

    def __init__(self, path):
        self.path = path
        if not os.path.exists(path):
            os.mkdir(path)
            os.chmod(path, 0700)

        if not os.path.isdir(path):
            raise IOError("Cannot create configuration directory '%' because a file with the same name exists already."%path)

        if not os.path.exists(self.release_dump_path):
            os.mkdir(self.release_dump_path)
            os.chmod(self.release_dump_path, 0700)

        if not os.path.isdir(self.release_dump_path):
            raise IOError("Cannot create directory '%' because a file with the same name exists already."%self.release_dump_path)
            
        self.db_engine = create_engine('sqlite:///%s/batzenca.db'%path, echo=False)

        Base.metadata.create_all(self.db_engine)

        self.db_session =  sessionmaker(bind=self.db_engine)()
        self.db_session.commit()

        from gnupg import GnuPG
        self.gnupg = GnuPG(path + os.path.sep + "gnupg")

        try:
            git.Repo(self.path)
        except git.InvalidGitRepositoryError:
            repo = git.Repo.init(self.path)
            # git ignore
            open(os.path.join(repo.working_dir, ".gitignore"), "w").write(Session._gitignore)
            # add useful files
            repo.git.add("batzenca.db")
            for filename in ("pubring.gpg", "secring.gpg", "trustdb.gpg"):
                fullpath = os.path.join(repo.working_dir, "gnupg", filename)
                if os.path.exists(fullpath):
                    repo.git.add(fullpath)
            repo.git.commit(m="initial commit")
            
    def commit(self, snapshot=False, *args, **kwds):
        self.db_session.commit()
        if snapshot:
            self.snapshot(*args, **kwds)

    @property
    def query(self):
        return self.db_session.query

    @property
    def add(self):
        return self.db_session.add

    @property
    def add_all(self):
        return self.db_session.add_all

    @property
    def release_dump_path(self):
        """Path for storing releases in .asc and .yaml format."""
        return os.path.join(self.path, "releases")

    def snapshot(self, verbose=False, msg="snapshot"):
        repo = git.Repo(self.path)
        repo.git.add("batzenca.db")
        repo.git.add("gnupg")
        if verbose:
            print repo.git.status()
        if repo.is_dirty():
            out = repo.git.commit(m=msg)
            if verbose:
                print out
        
BATZENCADIR  = os.environ.get("BATZENCADIR", os.path.expanduser("~") + os.path.sep + ".batzenca")
session = Session(BATZENCADIR)




"""
.. module:: utils
 
.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Various utility functions.
"""

import sys

def thunderbird_rules(release, mime_encode=False, mime_filename=None):
    """Return an XML string which matches that used by Thunderbird/Icedove to
    store "per-recipient rules"

    :param batzenca.database.releases.Release release: the target release

    :param boolean mime_encode: MIME encode before outputting

    :param str mime_filename: filename used for MIME encoding (default:
        '<mailinglist.name>_rules_<year>-<month>-<day>.xml')

    """
    import StringIO
    fh = StringIO.StringIO()

    fh.write("""<?xml version="1.0" ?>\n""")
    fh.write("  <pgpRuleList>\n")
    key_ids = [str(key.kid) for key in release.active_keys]
    fh.write("""    <pgpRule email="{%s}" encrypt="2" keyId="%s" negateRule="0" pgpMime="2" sign="2"/>\n"""%(release.mailinglist.email,", ".join(key_ids)))
    fh.write("  </pgpRuleList>\n")
    content = fh.getvalue()
    fh.close()

    if mime_encode:
        from email import encoders
        from email.mime.base import MIMEBase

        xml_rules = MIMEBase('application', 'xml')
        xml_rules.set_payload(content)
        encoders.encode_base64(xml_rules)

        if mime_filename is None:
            mime_filename = "%s_rules_%04d%02d%02d.xml"%(release.mailinglist.name,
                                                         release.date.year, release.date.month, release.date.day)
        
        xml_rules.add_header('Content-Disposition', 'attachment', filename=mime_filename)
        return xml_rules
    else:
        return content

    
def plot_nkeys(mailinglists, active_only=True):
    """Write a PDF file which plots the number of (active) keys over time in all releases for all ``mailinglists``.

    :param iterable mailinglists: instances of :class:`batzenca.database.mailinglists.MailingList`

    :param boolean active_only: only consider active keys; if ``False`` all keys are considered


    .. note:: requires matplotlib
    """
    import matplotlib.dates as mdates
    import matplotlib.pyplot as plt

    plt.clf()
    plt.gca().xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%Y'))
    for mailinglist in mailinglists:
        x = [release.date for release in mailinglist.releases]
        if active_only:
            y = [len(release.active_keys) for release in mailinglist.releases]
        else:
            y = [len(release.keys) for release in mailinglist.releases]

        plt.plot(x,y, linewidth=2.5, alpha=0.9, marker='o', label=mailinglist.name)

    plt.legend(loc='upper left')
    plt.gcf().set_size_inches(20,5) 
    plt.gcf().autofmt_xdate()
    plt.savefig("-".join([mailinglist.name  for mailinglist in mailinglists]) + ".pdf")

def find_orphaned_keys():
    """Find keys in the GnuPG database which do not have an instance of :class:`Key` associated with it.

    This library uses two databases: the GnuPG database of keys and the database storing
    metainformation which the user mostly works with. This function returns those keys in the GnuPG
    database which do not have an entry in the user facing database.
    """
    from batzenca import EntryNotFound, Key, session
    orphans = []
    for key in session.gnupg.ctx.op_keylist_all(None, 0):
        dbkey = None
        for sk in key.subkeys:
            try:            
                dbkey = Key.from_keyid(int(sk.keyid,16))
                break
            except EntryNotFound:
                pass
        if dbkey is None:
            orphans.append(Key(int(key.subkeys[0].keyid,16)))
    return tuple(orphans)

def import_new_key(key, peer=None, mailinglists=None, force=False):
    """Import a new ``key`` for ``peer``.

    This function does the following for all mailing lists on which the provided
    ``peer`` is currently subscribed.

    1. check if the key passes policy checks

    2. revoke all signatures on the old key of the provided peer which is
       replaced by the provided ``key``

    3. sign the key with the CA key

    4. create a new release if necessary

    5. add the key the current release

    6. delete all superfluous signatures
    
    :param batzenca.database.keys.Key key: the new key

    :param batzenca.database.peers.Peer peer: the peer to use in case it cannot
        be determined automtically

    :param iterable mailinglists: a list of mailing lists to consider or
        ``None`` for all.

    :param boolean force: by default a key is only added to a list if the
        matching peer has a key in the current release. If ``force==True`` the
        key is added unconditionally.

    """
    from batzenca.database import MailingList, Peer
    
    if peer is None:
        peer = Peer.from_email(key.email)
    if peer is None:
        raise ValueError("No peer provided")

    print " key:: %s"%key
    print "peer:: %s"%peer
    print

    if mailinglists is None:
        mailinglists = MailingList.all()

    mailinglists = [m for m in mailinglists if force or (peer in m.current_release)]
        
    # 1. check
            
    for mailinglist in mailinglists:
        if not mailinglist.policy.check(key, check_ca_signature=False):
            raise ValueError("key %s does not pass policy check for %s"%(key, mailinglist))
    
    # 2. update peer

    if peer.key and peer.key != key:
        for mailinglist in mailinglists:
            if mailinglist.policy.ca in peer.key.signatures:
                peer.key.revoke_signature(mailinglist.policy.ca)
    key.peer = peer

    # 5. add the key the current release

    signatures = set([key])
    
    for mailinglist in mailinglists:
        print "#",mailinglist,"#"

        key.sign(mailinglist.policy.ca)
        signatures.add(mailinglist.policy.ca)
        
        if mailinglist.current_release.published:
            _ = mailinglist.new_release()

        mailinglist.current_release.deactivate_invalid()
        if key not in mailinglist.current_release:
            mailinglist.current_release.add_key(key)

        print "done"
        print

    # 6. delete all superfluous signatures
    key.clean(signatures)

    print "signatures:"
    for signature in key.signatures:
        print "-",signature

def smtpserverize(email):
    """Read BATZENCA_DIR/smtp.cfg and construct smtpserver object for entry matching ``email``

    :param string email: an e-mail address

    It is assumed that smtp.cfg has the following form::

        [email]
        host: example.com
        port: 25
        username: user
        password: secret!
        security: starttls 

    Supported values for security are "starttls" and "tls".

    """
    import ConfigParser
    import smtplib
    import os
    from batzenca import session
    
    config = ConfigParser.ConfigParser()
    config.read(os.path.join(session.path, "smtp.cfg"))
    host = config.get(email, 'host')
    port = config.getint(email, 'port')
    security = config.get(email, 'security')

    if security.lower() == 'starttls':
        smtpserver = smtplib.SMTP(host, port=port)
        smtpserver.ehlo(name='localhost')
        smtpserver.starttls()
        smtpserver.login(config.get(email, "username"), config.get(email, "password"))
    elif security.lower() == 'tls':
        smtpserver = smtplib.SMTP_SSL(server, port=port)
        smtpserver.ehlo(name='localhost')
        smtpserver.login(config.get(email, "username"), config.get(email, "password"))
    else:
        raise ValueError("value '%s' for security not understood. Supported options are 'starttls' and 'tls'."%security)
    return smtpserver

def publish(mailinglists=None, debug=False, msg="", include_thunderbird_rules=True):
    """Publish all outstanding releases.

    :param iterable mailinglists: a list of mailing lists to consider or ``None`` for all.

    :param boolean debug: do not send e-mail to lists but to CA e-mail address.

    :param string msg: message to be included in git commit

    :param boolean include_thunderbird_rules: attach XML formated rules for Thunderbird/Icedove

    """
    from batzenca import MailingList, session
    
    if mailinglists is None:
        mailinglists = MailingList.all()
    
    published_releases = []
    smtpservers = {}

    for i, mailinglist in enumerate(mailinglists):
        print "%3d. [%s]"%(i, mailinglist),
        release = mailinglist.current_release

        if release.published:
            print
            continue

        if include_thunderbird_rules:
            attachments = (thunderbird_rules(release, mime_encode=True),)
        else:
            attachments = tuple()

        email = release.policy.ca.email
        
        if email in smtpservers:
            smtpserver = smtpservers[email]
        else:
            smtpserver = smtpserverize(email)
            smtpservers[email] = smtpserver

        release.send(smtpserver, debug=debug, attachments=attachments)
        
        published_releases.append(mailinglist.name)

        print "published"
        sys.stdout.flush()

    for smtpserver in smtpservers.itervalues():
        smtpserver.quit()

    msg = msg + " " + ", ".join(published_releases)
    session.commit(verbose=True, snapshot=True, msg=msg)
        


from database import EntryNotFound
from database import Key
from database import Peer, merge_peers
from database import MailingList
from database import Policy, PolicyViolation
from database import Release

import datetime

from batzenca.session import session

import warnings
warnings.simplefilter("always", PolicyViolation)

from batzenca.util import import_new_key


"""
.. module:: pgpmime

.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Realises PGP/MIME signatures and encryption.

.. note::

    This module is based on https://pypi.python.org/pypi/pgp-mime/
"""
from email import Message
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.encoders import encode_7or8bit
from email import encoders
from email.mime.base import MIMEBase
import email

import os
import mimetypes

from batzenca.session import session


class PGPMIMEsigned(MIMEMultipart):
    """
    A MIME-type for PGP/MIME signed messages.

    :param msg: a MIME object
    :param batzenca.database.keys.Key signer: the signing key
    """
    def __init__(self, msg=None, signer=None):
        if msg is None:
            return
        if msg.is_multipart():
            # we need these to get our message correctly parsed by KMail and Thunderbird
            msg.preamble = 'This is a multi-part message in MIME format.'
            msg.epilogue = '' 

        if signer is not None:
            msg_str = flatten(msg)
            sig     = session.gnupg.msg_sign(msg_str, signer.kid)
            sig = MIMEApplication(_data=sig,
                                  _subtype='pgp-signature; name="signature.asc"',
                                  _encoder=encode_7or8bit)
            sig['Content-Description'] = 'This is a digital signature.'
            sig.set_charset('us-ascii')

            MIMEMultipart.__init__(self, 'signed', micalg='pgp-sha1', protocol='application/pgp-signature')
            self.attach(msg)
            self.attach(sig)
            
    @classmethod
    def from_parts(cls, msg, sig):
        """Assemble a PGP/MIME signed message from its two parts: the message and the signature,
        both of which already properly encoded.

        :param msg: a MIME encoded message :param sig: a PGP/MIME signature

        """
        self = PGPMIMEsigned()
        MIMEMultipart.__init__(self, 'signed', micalg='pgp-sha1', protocol='application/pgp-signature')
        self.attach(msg)
        self.attach(sig)
        return self
            
    @property
    def signatures(self):
        """Return a list of keys for which this message has valid signatures.

        :return: a tuple of :class:`batzenca.database.keys.Key` objects
        """
        from batzenca.database.keys import Key
        
        subparts = self.get_payload()
        assert(len(subparts) == 2)
        msg, sig = subparts
        msg_str = flatten(msg)
        sigs = session.gnupg.sig_verify(msg_str, sig.get_payload())
        res = []
        for sig in sigs:
            try:
                key = Key.from_keyid(int(sig[-16:],16))
            except EntryNotFound:
                key = Key(int(sig[-16:],16))
            res.append(key)
        return tuple(res)

    def is_signed_by(self, signer):
        """Return ``True`` if the message is signed by ``signer``

        :param :class:`batzenca.database.keys.Key` signer: the potential signing key

        """
        from batzenca import EntryNotFound, Key

        signatures = self.signatures
            
        for sig in signatures:
            if isinstance(signer, Key):
                if key == signer:
                    return True
            else:
                if key.email == signer:
                    return True
        return False
        
class PGPMIMEencrypted(MIMEMultipart):
    """
    A MIME-type for PGP/MIME encrypted messages.

    :param msg: a MIME object
    :param iterable recipients: an iterable of recipients, where each entry is a
        :class:`batzenca.database.keys.Key` object.
    """
    def __init__(self, msg, recipients):
        MIMEMultipart.__init__(self, 'encrypted', micalg='pgp-sha1', protocol='application/pgp-encrypted')

        body = flatten(msg)
        encrypted = session.gnupg.msg_encrypt(body, [r.kid for r in recipients])

        payload = MIMEApplication(_data=encrypted,
                                  _subtype='octet-stream',
                                  _encoder=encode_7or8bit)
        payload['Content-Disposition'] = 'inline; name="encrypted.asc"'
        payload.set_charset('us-ascii')

        control = MIMEApplication(_data='Version: 1\n',
                                  _subtype='pgp-encrypted',
                                  _encoder=encode_7or8bit)
        control.set_charset('us-ascii')
        
        self.attach(control)
        self.attach(payload)
        self['Content-Disposition'] = 'attachment'

    def decrypt(self):
        """
        Decrypt this encrypted PGP/MIME message.

        :return: a MIMEType object
        """
        subparts = self.get_payload()
        assert(len(subparts) == 2)
        control, payload = subparts

        payload = payload.get_payload()
        raw = session.gnupg.msg_decrypt(payload)
        msg = email.message_from_string(raw)

        if msg.is_multipart():
            subparts = msg.get_payload()
            if len(subparts) == 2:
                msg, sig = subparts
                if "pgp-signature" in sig.get_content_subtype():
                    return PGPMIMEsigned.from_parts(msg, sig)
        return msg
        
def PGPMIME(msg, recipients, signer):
    """Construct a PGP/MIME signed and encrypted message from MIME message ``msg``, signed by
    ``signer`` and encrypted for all entries of ``recipients``.

    :param msg: the message to be signed and encrypted
    :param iterable recipients: an iterable of recipients, where each entry is a
        :class:`batzenca.database.keys.Key` object.
    :param signer: a signing key as an object of type :class:`batzenca.database.keys.Key`

    :return: a PGP/MIME encrypted message, containing a PGP/MIME signed message.
    """
    return PGPMIMEencrypted( PGPMIMEsigned(msg, signer), recipients)

def flatten(msg):
    from cStringIO import StringIO
    from email.generator import Generator
    fp = StringIO()
    g = Generator(fp, mangle_from_=False)
    g.flatten(msg)
    text = fp.getvalue()

    return '\r\n'.join(text.splitlines()) + '\r\n'



"""
.. module:: mailinglists
 
.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Mailing lists is what we do the work for.
"""

from sqlalchemy import Column, Integer, String, Date, ForeignKey, UnicodeText
from sqlalchemy.orm import relationship

from base import Base, EntryNotFound
from releases import Release, ReleaseKeyAssociation
from keys import Key
from peers import Peer

import warnings

class MailingList(Base):
    """This class represents a mailing list. A mailing list is an object which has one or more
    :class:`batzenca.database.releases.Release` objects associationed with it.

    :param str name: the name of this mailing list.
    :param str email: the e-mail address of this mailing list.

    :param policy: an object of type :class:`batzenca.database.policies.Policy` which is applied by
        default to releases

    :param str description: an informal description of this mailing list.

    :param str new_member_msg: this message ought to be sent to peers when they join the list for
      the first time. This message could contain etiquette etc.  This is string supports a limited
      number of template fields which are replaced by the actual values in
      :func:`batzenca.database.releases.Release.welcome_messages`. These fields are:

        ``{peer}`` - the name of the peer

        ``{mailinglist}`` - the name of this mailing list

        ``{mailinglist_email}`` - the email address of the mailing list on which the user is
        subscribed

        ``{ca}`` - the CA's name

        ``{ca_email}`` - the CA's e-mail address
    
    :param str key_update_msg: this message is sent when a new release is published. It typically
      contains a list of all keys associated with this release. For this reason this string supports
      a limited number of template fields which are replaced by the actual values in
      :func:`batzenca.database.releases.Release.publish`. These fields are:
        
        ``{mailinglist}`` - the name of this mailing list
        
        ``{peers_in}`` - a comma-separated list of names
        (:class:`batzenca.database.peers.Peer.name`) which are new in the current release.
        
        ``{peers_changed}`` - a comma-separated list of names
        (:class:`batzenca.database.peers.Peer.name`) which are changed their keys in the current
        release.
        
        ``{peers_out}`` - a comma-separated list of names
        (:class:`batzenca.database.peers.Peer.name`) which left the mailing list in the current
        release.
        
        ``{keys_in}`` - a line-break-separated list of keys (printed using
        :func:`batzenca.database.releases.Release._format_entry`) that joined the mailing list in
        the current release.
        
        ``{keys_out}`` - a line-break-separated list of keys (printed using
        :func:`batzenca.database.releases.Release._format_entry`) that left this mailing list in the
        current release.
        
        ``{keys}`` - a complete line-break-separated list of keys (printed using
        :func:`batzenca.database.releases.Release._format_entry`) in the current release.

        ``{ca}`` - the CA's name

        ``{ca_email}`` - the CA's e-mail address

        ``{dead_man_switch}`` - an optional message stating that no attempts were made to compel the
        CA to compromise the integrity of the list.
    
    :param str key_expiry_warning_msg: this message is sent when a key is about to expire, to warn
      the user about this fact. This is string supports a limited number of template fields which
      are replaced by the actual values in
      :func:`batzenca.database.releases.Release.key_expiry_message`. These fields are:

        ``{peer}`` - the name of the peer

        ``{keyid}`` - the key id of the expiring key

        ``{expiry_date}`` - the expiry date

        ``{mailinglist}`` - the name of the mailing list on which the user is subscribed

        ``{mailinglist_email}`` - the email address of the mailing list on which the user is
        subscribed

        ``{ca_email}`` - the CA's e-mail address

    :param str dead_man_switch_msg: this message is included in
        :attr:`batzenca.database.mailinglists.MailingList.key_update_msg` by
        :func:`batzenca.database.releases.Release.__call__` if the parameter ``still_alive`` is set
        to ``True`` when calling it.

    """

    __tablename__ = 'mailinglists'

    id          = Column(Integer, primary_key=True) 
    name        = Column(String, unique=True, nullable=False)
    email       = Column(String)
    description = Column(UnicodeText)
    new_member_msg         = Column(UnicodeText) # the message we send to a new member
    key_update_msg         = Column(UnicodeText)
    key_expiry_warning_msg = Column(UnicodeText)
    dead_man_switch_msg    = Column(UnicodeText)

    policy_id   = Column(Integer, ForeignKey('policies.id'))
    policy      = relationship("Policy")

    def __init__(self, name, email, policy=None, description='', new_member_msg='', key_update_msg='', key_expiry_warning_msg='', dead_man_switch_msg=''):
        self.name = name
        self.email = email
        self.policy = policy
        self.description = unicode(description)
        self.key_update_msg = unicode(key_update_msg)
        self.new_member_msg = unicode(new_member_msg)
        self.key_expiry_warning_msg = unicode(key_expiry_warning_msg)
        self.dead_man_switch_msg    = unicode(dead_man_switch_msg)

    @classmethod
    def from_name(cls, name):
        from batzenca.session import session
        res = session.db_session.query(cls).filter(cls.name == name)
        if res.count() == 0:
            raise EntryNotFound("No mailinglist with name '%s' found in database."%name)
        else:
            if res.count() > 1:
                warnings.warn("More than one mailinglist with name '%s' found, picking first one."%name)
            return res.first()

    @classmethod
    def all(cls):
        """Return all mailing lists from the database"""
        from batzenca import session
        res = session.db_session.query(MailingList)
        return tuple(res.all())
        
    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)

    @property
    def current_release(self):
        """
        Return the current release.
        """
        try:
            return self.releases[-1]
        except IndexError:
            raise IndexError("The mailing list '%s' has no release yet."%self)
    
    def new_release(self, date=None, inherit=True, deactivate_invalid=True, delete_old_inactive_keys=5):
        """Create a new release for this mailing list.

        If ``inherit == True`` then :func:`batzenca.database.releases.Release.inherit` is called and
        the remaining parameters are passed through. Otherwise a new empty release is created and no
        parameter except for ``date`` would make sense.
        
        :param date: the date of this release. If ``None`` today's date is used
        :param boolean inherit: inherit keys and policy from the currently active release
        :param boolean deactivate_invalid: deactivate keys which are no longer valid, e.g. because
          they are expired. Only applies if ``inherit=True``
        :param boolean delete_old_inactive_keys: delete inactive keys which have been around for a
          while, this parameter is passed to
          :func:`batzenca.database.releases.Release.delete_old_inactive_keys` as ``releasecount``

        .. note::

           If ``inherit == True`` the returned object was added to the main session and is
           associated with this mailing list.

        """
        if inherit is True:
            return self.current_release.inherit(date=date, deactivate_invalid=deactivate_invalid, delete_old_inactive_keys=delete_old_inactive_keys)
        elif inherit:
            if inherit.mailinglist is not self:
                raise ValueError("Cannot inherit from release '%s' because it is for '%s' instead of '%s'."%(inherit, inherit.mailinglist, self))
            return inherit.inherit(date=date, deactivate_invalid=deactivate_invalid, delete_old_inactive_keys=delete_old_inactive_keys)
        else:
            return Release(mailinglist=self, date=date, active_keys=[], inactive_keys=[], policy=self.policy)

    def __contains__(self, obj):
        """Return ``True`` if ``obj`` was ever in any release for this mailing list.

        :param obj: either an object of type :class:`batzenca.database.peers.Peer` or
          :class:`batzenca.database.keys.Key`

        """
        from batzenca.session import session
        from sqlalchemy.sql import or_
        
        query = session.db_session.query(Release).filter(Release.mailinglist == self)
        query = query.join(ReleaseKeyAssociation).filter(ReleaseKeyAssociation.right_id == Release.id)
        query = query.join(Key)

        if isinstance(obj, Peer):
            q = [ReleaseKeyAssociation.left_id == key.id for key in obj.keys]
            query = query.filter(or_(*q))
        elif isinstance(obj, Key):
            query = query.filter(ReleaseKeyAssociation.left_id == obj.id)
        else:
            raise TypeError("Object of type '%s' passed, but only types `Peer` and `Key` are supported."%(type(obj)))
            
        if query.first() is not None:
            return True
        else:
            return False
            


#!/usr/bin/python
"""
Mon signed Leia's and Luke's key but not Han's
"""

import batzenca
import batzenca.gnupg
import datetime
import os
import unittest
 
class TestGnuPGQuery(unittest.TestCase):
    def setUp(self):
        gnupghome = os.environ["GNUPGHOME"] = os.path.sep.join([os.getcwd(),"tests","batzencadir","gnupg"])
        self.gnupg = batzenca.gnupg.GnuPG(gnupghome)
        
        self.leia = "4E2584FC19840E5F"
        self.han  = "DB22248C8F4C0C37"
        self.luke = int("E843C898AB0FA2FD", 16)
        self.mon  = "FABA3916FB56A97D"
        
    def test_key_get(self):
        self.assertIsNotNone( self.gnupg.key_get(self.leia) )
        self.assertIsNotNone( self.gnupg.key_get(self.han ) )
        self.assertIsNotNone( self.gnupg.key_get(self.luke) )
        self.assertIsNotNone( self.gnupg.key_get(self.mon ) )
        with self.assertRaises(batzenca.gnupg.KeyError):
            self.gnupg.key_get( "0000000000000000" )

    def test_have_secret_key(self):
        self.assertFalse( self.gnupg.have_secret_key(self.leia) )
        self.assertFalse( self.gnupg.have_secret_key(self.han ) )
        self.assertFalse( self.gnupg.have_secret_key(self.luke) )

        self.assertTrue( self.gnupg.have_secret_key(self.mon) )

        self.assertFalse( self.gnupg.have_secret_key(0) )

    def test_key_uid(self):
        name, email, comment = self.gnupg.key_uid(self.mon)
        self.assertEqual(name, "Mon Mothma")
        self.assertEqual(email, "mon@batzen.ca")
        self.assertEqual(comment, "CA")

    def test_key_okay(self):
        self.assertTrue( self.gnupg.key_okay(self.leia) )
        self.assertTrue( self.gnupg.key_okay(self.han ) )
        self.assertTrue( self.gnupg.key_okay(self.luke) )
        self.assertTrue( self.gnupg.key_okay(self.mon ) )

    def test_key_validity(self):
        self.assertEqual( self.gnupg.key_validity(self.leia), 4 )
        self.assertEqual( self.gnupg.key_validity(self.han ), 0 )
        self.assertEqual( self.gnupg.key_validity(self.luke), 4 )
        self.assertEqual( self.gnupg.key_validity(self.mon ), 5 )

    def test_key_pubkey_algos(self):
        self.assertEqual( self.gnupg.key_pubkey_algos(self.leia), ( 1,  1) )
        self.assertEqual( self.gnupg.key_pubkey_algos(self.han ), (17, 16) )
        self.assertEqual( self.gnupg.key_pubkey_algos(self.luke), ( 1,  1) )
        self.assertEqual( self.gnupg.key_pubkey_algos(self.mon ), ( 1,  1) )

    def test_key_expires(self):
        self.assertEqual( self.gnupg.key_expires(self.leia), False )
        self.assertEqual( self.gnupg.key_expires(self.han ), datetime.date(2015, 11, 17) )
        self.assertEqual( self.gnupg.key_expires(self.luke), datetime.date(2015,  5, 11) )
        self.assertEqual( self.gnupg.key_expires(self.mon ), False )

    def test_key_timestamp(self):
        self.assertEqual( self.gnupg.key_timestamp(self.leia), datetime.date(2013, 11, 17) )
        self.assertEqual( self.gnupg.key_timestamp(self.han ), datetime.date(2013, 11, 17) )
        self.assertEqual( self.gnupg.key_timestamp(self.luke), datetime.date(2013, 11, 17) )
        self.assertEqual( self.gnupg.key_timestamp(self.mon ), datetime.date(2013, 11, 17) )

    def test_key_min_len(self):
        self.assertEqual( self.gnupg.key_min_len(self.leia), 2048 )
        self.assertEqual( self.gnupg.key_min_len(self.han ), 1024 )
        self.assertEqual( self.gnupg.key_min_len(self.luke), 4096 )
        self.assertEqual( self.gnupg.key_min_len(self.mon ), 2048 )
    
    def test_key_any_uid_is_signed_by(self):
        self.assertTrue( self.gnupg.key_any_uid_is_signed_by(self.leia, self.mon) )
        self.assertTrue( self.gnupg.key_any_uid_is_signed_by(self.luke, self.mon) )
        self.assertTrue( self.gnupg.key_any_uid_is_signed_by(self.mon,  self.mon) )

        self.assertFalse( self.gnupg.key_any_uid_is_signed_by(self.han, self.mon) )

        self.assertFalse( self.gnupg.key_any_uid_is_signed_by(self.leia, self.han) )
        self.assertFalse( self.gnupg.key_any_uid_is_signed_by(self.luke, self.han) )
        self.assertFalse( self.gnupg.key_any_uid_is_signed_by(self.mon,  self.han) )

        self.assertTrue( self.gnupg.key_any_uid_is_signed_by(self.han, self.han) )

    def test_key_signatures(self):
        self.assertEqual( self.gnupg.key_signatures(self.leia), set(['4E2584FC19840E5F', 'FABA3916FB56A97D']) )
        self.assertEqual( self.gnupg.key_signatures(self.han ), set(['DB22248C8F4C0C37']) )
        self.assertEqual( self.gnupg.key_signatures(self.luke), set(['FABA3916FB56A97D', 'E843C898AB0FA2FD']) )
        self.assertEqual( self.gnupg.key_signatures(self.mon ), set(['FABA3916FB56A97D']) )

    def test_key_okay_encrypt(self):
        self.assertTrue(  self.gnupg.key_okay_encrypt(self.leia) )
        self.assertFalse( self.gnupg.key_okay_encrypt(self.han ) )
        self.assertTrue(  self.gnupg.key_okay_encrypt(self.luke) )
        self.assertTrue(  self.gnupg.key_okay_encrypt(self.mon ) )

    def test_keys_export(self):
        keys = self.gnupg.keys_export( [self.leia, self.luke] )
        self.assertTrue(keys.startswith('-----BEGIN PGP PUBLIC KEY BLOCK-----\n'))
        self.assertTrue(keys.endswith('-----END PGP PUBLIC KEY BLOCK-----\n'))

    def test_msg_sign(self):
        msg = "A long time ago in a galaxy far, far away"
        sig = self.gnupg.msg_sign(msg, self.mon)
        self.assertTrue(sig.startswith('-----BEGIN PGP SIGNATURE-----\n'))

        with self.assertRaises(ValueError):
            signature = self.gnupg.msg_sign(msg, self.leia)
        
    def test_msg_encrypt(self):
        msg = "A long time ago in a galaxy far, far away"
        ciphertext = self.gnupg.msg_encrypt(msg, [self.leia, self.luke])
        self.assertTrue(ciphertext.startswith('-----BEGIN PGP MESSAGE-----\n'))

        with self.assertRaises(ValueError):
            ciphertext = self.gnupg.msg_encrypt(msg, [self.han, self.mon])

        ciphertext = self.gnupg.msg_encrypt(msg, [self.han, self.mon], always_trust=True)
        self.assertTrue(ciphertext.startswith('-----BEGIN PGP MESSAGE-----\n'))

    def test_msg_decrypt(self):
        msg = "A long time ago in a galaxy far, far away"
        ciphertext = self.gnupg.msg_encrypt(msg, [self.leia, self.mon])
        msg2 = self.gnupg.msg_decrypt(ciphertext)
        self.assertEqual(msg, msg2)

        msg = "A long time ago in a galaxy far, far away"
        ciphertext = self.gnupg.msg_encrypt(msg, [self.leia])

        with self.assertRaises(ValueError):
            msg2 = self.gnupg.msg_decrypt(ciphertext)

    def test_sig_verify(self):
        msg = "A long time ago in a galaxy far, far away"
        sig = self.gnupg.msg_sign(msg, self.mon)
        fprs = self.gnupg.sig_verify(msg, sig)
        self.assertEqual(len(fprs), 1)
        self.assertEqual(self.mon, fprs[0][-16:])

if __name__ == '__main__':
    unittest.main()


"""
.. module:: gnupg

.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Interface to low(er) level GnuPG interfaces.
"""

import pyme
import pyme.core
import pyme.pygpgme
import pyme.errors
import pyme.constants.keylist
import pyme.constants.sig
import pyme.constants.status

from collections import namedtuple
UID = namedtuple('UID', ['name', 'email',  'comment'])

import os
import datetime

class KeyError(Exception):
    """
    We raise this exception if there is a problem with a key
    """
    pass

class GnuPG(object):
    """A GnuPG context.

    :param str home_dir: specifiy ``GNUPGHOME``, if ``None`` an implementation default is used.

    """
    GPGME_PK_RSA   = pyme.pygpgme.GPGME_PK_RSA
    GPGME_PK_RSA_E = pyme.pygpgme.GPGME_PK_RSA_E
    GPGME_PK_RSA_S = pyme.pygpgme.GPGME_PK_RSA_S
    GPGME_PK_ELG_E = pyme.pygpgme.GPGME_PK_ELG_E
    GPGME_PK_DSA   = pyme.pygpgme.GPGME_PK_DSA
    GPGME_PK_ELG   = pyme.pygpgme.GPGME_PK_ELG

    alg_to_str = { GPGME_PK_RSA  : "GPGME_PK_RSA",
                   GPGME_PK_RSA_E: "GPGME_PK_RSA_E",
                   GPGME_PK_RSA_S: "GPGME_PK_RSA_S",
                   GPGME_PK_ELG_E: "GPGME_PK_ELG_E",
                   GPGME_PK_DSA  : "GPGME_PK_DSA",
                   GPGME_PK_ELG  : "GPGME_PK_ELG" }

    str_to_alg = { "GPGME_PK_RSA"  : GPGME_PK_RSA,
                   "GPGME_PK_RSA_E": GPGME_PK_RSA_E,
                   "GPGME_PK_RSA_S": GPGME_PK_RSA_S,
                   "GPGME_PK_ELG_E": GPGME_PK_ELG_E,
                   "GPGME_PK_DSA"  : GPGME_PK_DSA,
                   "GPGME_PK_ELG"  : GPGME_PK_ELG }

    _key_cache = {}

    @staticmethod
    def is_active(subkey):
        """
        Return ``True`` if ``subkey`` is neither revoked, expired or disabled.

        :param subkey:  a PyGPGMe subkey object

        """
        return not (subkey.revoked or subkey.expired or subkey.disabled)

    def __init__(self, home_dir=None):
        pyme.core.check_version(None)

        self._home_dir = home_dir

        if home_dir is not None:
            if not os.path.exists(home_dir):
                os.mkdir(home_dir)
                os.chmod(home_dir, 0700)

            for engine in pyme.core.get_engine_info():
                pyme.core.set_engine_info(engine.protocol, engine.file_name, home_dir)

        self.ctx = pyme.core.Context()
        self.ctx.set_keylist_mode(pyme.constants.keylist.mode.SIGS)
        self.ctx.set_armor(1)

    @property
    def home_dir(self):
        return str(self._home_dir)

    def key_get(self, keyid):
        """
        Get the key object matching ``keyid``.

        :param keyid: a 16 character string encoding an integer in hexadecimal notation or an
            integer :math:`< 2^{64}`
        """
        if isinstance(keyid, pyme.pygpgme._gpgme_key):
            return keyid

        # we are caching for performance reasons
        # TODO: does anything depend on us not caching?
        if keyid in self._key_cache:
            return self._key_cache[keyid]

        try:
            try:
                key = self.ctx.get_key("0x%x"%keyid, 0)
                self._key_cache[keyid] = key
                return key
            except TypeError:
                pass
            try:
                if keyid.startswith("0x"):
                    key = self.ctx.get_key(str(keyid), 0)
                else:
                    key = self.ctx.get_key("0x"+str(keyid), 0)
                self._key_cache[keyid] = key
                return key
            except AttributeError:
                raise KeyError("Key '%s' not found."%keyid)
        except pyme.errors.GPGMEError, e:
            raise KeyError("Key '%s' not found."%keyid)

    def have_secret_key(self, keyid):
        """
        Return ``True`` if we have the secret key for the key identified by``keyid``.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.

        """
        try:
            try:
                key = self.ctx.get_key("0x%x"%keyid, True)
                if key:
                    return True
            except TypeError:
                pass
            try:
                if keyid.startswith("0x"):
                    key = self.ctx.get_key(str(keyid), True)
                else:
                    key = self.ctx.get_key("0x"+str(keyid), True)
                if key:
                    return True
            except AttributeError:
                return False
        except pyme.errors.GPGMEError, e:
            return False

    def key_uid(self, keyid):
        """
        Return a named tuple ``(name,email,comment)`` with the default UID of the key ``keyid``

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        uids = self.key_get(keyid).uids
        return UID(unicode(uids[0].name, 'utf-8'), unicode(uids[0].email, 'utf-8'), unicode(uids[0].comment, 'utf-8'))

    def key_okay(self, keyid):
        """
        A key is "okay" if it has at least one not-expired, not-disabled,
        not-revoked subkey for both signing and encrypting.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.

        """
        key = self.key_get(keyid)

        # the .invalid attribute is not used by GnuPG
        assert(all(subkey.invalid == 0 for subkey in key.subkeys))

        subkeys = [subkey for subkey in key.subkeys if GnuPG.is_active(subkey)]
        if len(subkeys) == 0:
            return False

        if not any(subkey.can_sign for subkey in subkeys) or not any(subkey.can_encrypt for subkey in subkeys):
            return False
        return True

    def key_validity(self, keyid):
        """
        Return the validity of the key ``keyid``.

        A key's validity expresses the level of certainty that a given key belongs to the claimed
        UID. Where 0 is no certainty, and 5 is ultimate certainty (you own the secret key yourself).

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        return max([uid.validity for uid in key.uids])

    def key_pubkey_algos(self, keyid):
        """Return a tuple of public-key enryption/signature algorithms for the key ``keyid``

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        return tuple([subkey.pubkey_algo for subkey in key.subkeys if GnuPG.is_active(subkey)])

    def key_expires(self, keyid):
        """
        Return the nearest date when any of the expiring encryption subkeys of
        ``keyid`` is going to expire.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)

        timestamps = [subkey.expires for subkey in key.subkeys if (subkey.expires and subkey.can_encrypt)]
        if timestamps:
            return datetime.date.fromtimestamp(min(timestamps))
        else:
            return False

    def key_expired(self, keyid):
        """
        Return ``True`` if the key ``keyid`` is expired.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        subkeys = [subkey for subkey in key.subkeys if (subkey.can_encrypt and not subkey.expired)]
        if len(subkeys) == 0:
            return True
        else:
            return False

    def key_timestamp(self, keyid):
        """
        Return the earliest timestamp on any of the subkeys of ``keyid``.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        timestamp = min([subkey.timestamp for subkey in key.subkeys])
        return datetime.date.fromtimestamp(timestamp)

    def key_min_len(self, keyid):
        """
        Return the minimum key length of any of the subkeys of ``keyid``.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        return min([subkey.length for subkey in key.subkeys])


    def key_any_uid_is_signed_by(self, keyid, signer_keyid):
        """
        Return ``True`` if any uid of the key ``keyid`` is signed by the key ``signer_keyid``.

        :param keyid: the signee, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        :param signer_keyid: the potential signer, see :func:`batzenca.gnupg.GnuPG.key_get` for
            accepted formats.

        """
        key = self.key_get(keyid)

        signer_key = self.key_get(signer_keyid)

        sign_keys = set()
        for subkey in signer_key.subkeys:
            if subkey.can_sign and not subkey.disabled and not subkey.revoked:
                sign_keys.add(subkey.keyid)

        if sign_keys.intersection(self.key_signatures(keyid)):
            return True
        else:
            return False

    def key_signatures(self, keyid):
        """
        Return the list of keyids which match keys that signed the key ``keyid``.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        sigs = set()
        for uid in key.uids:
            for sig in uid.signatures:
                if not sig.expired and not sig.revoked:
                    sigs.add(sig.keyid)

        for uid in key.uids:
            for sig in uid.signatures:
                if sig.revoked and sig.keyid in sigs:
                    sigs.remove(sig.keyid)

        return sigs

    def key_fingerprint(self, keyid):
        """
        Return the fingerprint.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        return key.subkeys[0].fpr


    def key_okay_encrypt(self, keyid):
        """
        Return ``True`` if the key ``keyid`` can be used for encryption.

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        key = self.key_get(keyid)
        if  not self.key_okay(key) or not self.key_validity(key) >= 4:
            return False
        return True

    def keys_export(self, keyids):
        """
        Return string containing keys in list ``keyids`` in ASCII armor format

        :param keyid: see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        """
        from pyme.core import Data
        export_keys = Data()
        keys = [self.key_get(keyid) for keyid in keyids]
        self.ctx.op_export_keys(keys, 0, export_keys)
        export_keys.seek(0,0)
        return export_keys.read()

    def msg_sign(self, msg, keyid):
        """
        Sign the message with the key

        :param str msg: the message
        :param keyid: the signing key, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted
            formats.
        """
        key = self.key_get(keyid)

        if not self.have_secret_key(keyid):
            raise ValueError("You do not have the secret key for %s in your GnuPG keyring."%keyid)

        msg = pyme.core.Data(msg)
        sig = pyme.core.Data()

        self.ctx.signers_clear()
        self.ctx.signers_add(key)

        self.ctx.op_sign(msg, sig, pyme.constants.sig.mode.DETACH)

        sig.seek(0,0)
        return sig.read()

    def msg_encrypt(self, msg, keyids, always_trust=False):
        """
        Encrypt the message under keys in the list ``keyids``.

        :param str msg: the message
        :param iterable keyids: a list of key ids, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats.
        :param boolean always_trust: If ``False`` on keys with validity :math:`>= 4` are accepted. Otherwise, any
          public key will do.
        """
        keys = [self.key_get(keyid) for keyid in keyids]

        if not always_trust:
            for key in keys:
                if  not self.key_okay(key) or not self.key_validity(key) >= 4:
                    raise ValueError("No UID of the key 0x%s has a sufficient level of validity, set always_trust=True if you want to force encryption."%key.subkeys[0].fpr[-16:])

        plain = pyme.core.Data(msg)
        cipher = pyme.core.Data()
        flags = pyme.constants.ENCRYPT_NO_ENCRYPT_TO
        if always_trust:
            flags |= pyme.constants.ENCRYPT_ALWAYS_TRUST
        retval = self.ctx.op_encrypt(keys, flags, plain, cipher)

        cipher.seek(0,0)
        return cipher.read()

    def msg_decrypt(self, cipher):
        """
        Decrypt a cyphertext.

        :param str cipher: the ciphertext
        """
        cipher = pyme.core.Data(cipher)
        plain  = pyme.core.Data()
        try:
            self.ctx.op_decrypt(cipher, plain)
            plain.seek(0,0)
            return plain.read()
        except pyme.errors.GPGMEError, msg:
            raise ValueError(msg)

    def sig_verify(self, msg, sig):
        """
        Return the list of keys - represented by fingerprint strings - that signed the message
        ``msg`` in the detached signature ``sig``.

        :param str msg: the message
        :param str sig: the signature
        """
        msg = pyme.core.Data(msg)
        sig = pyme.core.Data(sig)

        self.ctx.op_verify(sig, msg, None)
        result = self.ctx.op_verify_result()

        sigs = []
        for sign in result.signatures:
            if (sign.summary & pyme.constants.SIGSUM_VALID) == 1:
                sigs.append( sign.fpr )

        return tuple(sigs)

    def key_sign(self, keyid, signer_keyid, local=False):
        """
        Sign all user IDs of a key.

        :param keyid: the signee, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats
        :param signer_keyid: the signer, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats
        :param boolean local: if ``True`` a local signature (non exportable) is created
        """
        key = self.key_get(keyid)
        signer_key = self.key_get(signer_keyid)

        if not self.have_secret_key(signer_keyid):
            raise ValueError("You do not have the secret key for %s in your GnuPG keyring."%signer_keyid)

        out = pyme.core.Data()

        helper = {
            "GET_LINE"        : {"keyedit.prompt" : ("lsign" if local else "sign", "quit")},
            "GET_BOOL"        : {"sign_uid.okay" : "Y", "keyedit.save.okay" : "Y", "keyedit.sign_all.okay" : "Y"},
            "ALREADY_SIGNED"  : None,
            "GOT_IT"          : None,
            "NEED_PASSPHRASE" : None,
            "GOOD_PASSPHRASE" : None,
            "USERID_HINT"     : None,
            "EOF"             : None,
            "skip"            : 0,
            "data"            : out,
        }

        self.ctx.signers_clear()
        self.ctx.signers_add(signer_key)
        self.ctx.op_edit(key, edit_fnc, helper, out)
        self._key_cache = {} # invalidate the cache

    def key_delete_signature(self, keyid, signer_keyid):
        """
        Delete a specific signature from a key.

        This operation is the inverse of :func:`batzenca.gnupg.GnuPG.key_sign`.

        :param keyid: the signee, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats
        :param signer_keyid: the signer, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted
            formats

        """
        key = self.key_get(keyid)
        try:
            signer_key = self.key_get(signer_keyid)
        except KeyError:
            signer_key = signer_keyid

        out = pyme.core.Data()

        for i, uid in enumerate(key.uids):

            cleaner = {
                "GET_LINE"        : {"keyedit.prompt" : ("uid %d"%(i+1), "delsig", "save")},
                "GET_BOOL"        : {"keyedit.save.okay" : "Y", "keyedit.delsig.unknown" : "Y"},
                "GOT_IT"          : None,
                "NEED_PASSPHRASE" : None,
                "GOOD_PASSPHRASE" : None,
                "USERID_HINT"     : None,
                "EOF"             : None,

                "signer"          : signer_key,
                "skip"            : 0,
                "data"            : out,
            }
            self.ctx.op_edit(key, edit_fnc, cleaner, out)

        self._key_cache = {} # invalidate the cache

    def key_set_trust(self, keyid, trust):
        """
        Set the owner trust of a key.

        :param keyid: the signee, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats
        :param int trust: an integer between 0 and 5 (inclusive).

        """
        key = self.key_get(keyid)

        out = pyme.core.Data()

        helper = {
            "GET_LINE"        : { "keyedit.prompt" : ("trust", "quit"),
                                  "edit_ownertrust.value" : str(trust),
                              },
            "GET_BOOL"        : { "edit_ownertrust.set_ultimate.okay" : "Y" },
            "GOT_IT"          : None,
            "NEED_PASSPHRASE" : None,
            "GOOD_PASSPHRASE" : None,
            "USERID_HINT"     : None,
            "EOF"             : None,

            "skip"            : 0,
            "data"            : out,
        }
        self.ctx.op_edit(key, edit_fnc, helper, out)
        self._key_cache = {} # invalidate the cache

    def keys_import(self, data):
        """
        Import the keys from data.

        :param str data: PGP key data in ASCII format.

        """
        self._key_cache = {} # invalidate the cache
        data = pyme.core.Data(data)
        self.ctx.op_import(data)
        res =  self.ctx.op_import_result()
        return dict((r.fpr, r.status) for r in res.imports)

    def key_revsig(self, keyid, signer_keyid, code=4, msg=""):
        """
        Add a revocation signature for ``signer_keyid`` to ``keyid``.

        :param keyid: the signed key, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats
        :param signer_key: the key whose signature ought to be revoked, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats
        :param int code: an integer between 0 and 4 (inclusive) encoding the reason for the revocation
        :param str msg: an optional message added to the revocation

        .. note::

            This operation is not the inverse of :func:`batzenca.gnupg.GnuPG.key_sign`, but this
            operation adds an additional revocation signature to the key revoking a previous
            signature.
        """
        key = self.key_get(keyid)
        signer_key = self.key_get(signer_keyid)

        if not self.have_secret_key(signer_keyid):
            raise ValueError("You do not have the secret key for %s in your GnuPG keyring."%signer_keyid)

        out = pyme.core.Data()

        msg += "\n\n"
        msg = tuple(msg.splitlines())

        helper = {
            "GET_LINE"        : { "keyedit.prompt" : ("revsig", "quit"),
                                  "ask_revocation_reason.code" : str(code) ,
                                  "ask_revocation_reason.text" :  msg
                              },
            "GET_BOOL"        : { "ask_revoke_sig.okay" : "Y",
                                  "keyedit.save.okay" : "Y",
                                  "ask_revocation_reason.okay" : "Y" },
            "GOT_IT"          : None,
            "NEED_PASSPHRASE" : None,
            "GOOD_PASSPHRASE" : None,
            "USERID_HINT"     : None,
            "EOF"             : None,

            "signer"          : signer_key,
            "skip"            : 0,
            "data"            : out,
        }
        self.ctx.signers_clear()
        self.ctx.signers_add(signer_key)
        self.ctx.op_edit(key, edit_fnc, helper, out)
        self._key_cache = {} # invalidate the cache

    def key_edit(self, keyid):
        """
        Edit a key.

        :param keyid: key to edit, see :func:`batzenca.gnupg.GnuPG.key_get` for accepted formats

        .. warning:

           This will open an interactive session using rawinput
        """
        key = self.key_get(keyid)
        out = pyme.core.Data()

        helper = {
            "GOT_IT"          : None,
            "NEED_PASSPHRASE" : None,
            "GOOD_PASSPHRASE" : None,
            "EOF"             : None,

            "skip"            : 0,
            "data"            : out,
        }
        self.ctx.signers_clear()
        self.ctx.op_edit(key, edit_fnc, helper, out)
        self._key_cache = {} # invalidate the cache

# from pygpa

stat2str = {}
for name in dir(pyme.constants.status):
    if not name.startswith('__') and name != "util":
        stat2str[getattr(pyme.constants.status, name)] = name

def edit_fnc(stat, args, helper):
    try:
        while True:
            helper["data"].seek(helper["skip"],0)
            data = helper["data"].read()
            helper["skip"] += len(data)

            if stat2str[stat] in helper:
                if helper[stat2str[stat]] is None:
                     return ""

                if args in helper[stat2str[stat]]:
                    ret = helper[stat2str[stat]][args]
                    if isinstance(ret, tuple):
                        helper[stat2str[stat]][args] = ret[1:]
                        return ret[0]
                    return ret

            if stat2str[stat] == "GET_BOOL" and args == "keyedit.delsig.valid":
                try:
                    if any(sk.keyid[2:].upper() in data for sk in helper["signer"].subkeys):
                        return "Y"
                except AttributeError:
                    # for string signatures
                    if helper["signer"][2:].upper() in data:
                        return "Y"
                return "N"

            if stat2str[stat] == "GET_BOOL" and args == "ask_revoke_sig.one":
                for sk in helper["signer"].subkeys:
                    #0x....
                    if sk.keyid[2:].upper() in data:
                        return "Y"
                return "N"

            print data
            return raw_input("(%s) %s > " % (stat2str[stat], args))
    except EOFError:
        pass



"""
.. module:: policies

.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Policies specify conditions for keys to be part of releases.
"""

import sqlalchemy
from sqlalchemy import Column, Integer, String, Date, Boolean, ForeignKey, UnicodeText
from sqlalchemy.orm import relationship, backref

import datetime
import warnings

from base import Base, EntryNotFound
from keys import Key
from peers import Peer
from mailinglists import MailingList

class PolicyViolation(Warning):
    """This warning is issued a :class:`batzenca.database.keys.Key` violates the conditions set by a
    :class:`batzenca.database.policies.Policy`.

    The default behaviour for warnings of this kind is that a message is displayed whenever it
    occurs. This behaviour can be modified by using Python's built-in warnings module::

        >>> with warnings.catch_warnings():
        ...   warnings.simplefilter("ignore", PolicyViolation)
        ...   # do something

    Alternatively, these warnings can be turned into exceptions::

        >>> with warnings.catch_warnings():
        ...   warnings.simplefilter("error", PolicyViolation)
        ...   # do something
    
    """
    def __init__(self, msg):
        Warning.__init__(self, msg.encode('utf8'))

class Policy(Base):
    """Releases are checked against policies.

    :param str name: the name of this policy
    :param implementation_date: the date this policy was implemented
    :param batzenca.database.keys.Key ca: the CA key
    :param int key_len: the minimum required key length
    :param int key_lifespan: the maximal key lifespan in days. For example, if ``key_lifespan`` is
        365, then a key passes if it expires within the next 365 from the point in time when it is
        checked.
    :param iterable algorithms: tuple of allowed algorithms

    """
    __tablename__ = 'policies'

    id                  = Column(Integer, primary_key=True) #: database id
    name                = Column(String) #: the name of this policy
    implementation_date = Column(Date) #: the date this policy was implemented

    ca                  = relationship('Key') #: the CA key
    ca_id               = Column(Integer, ForeignKey('keys.id'), nullable=False)

    key_len             = Column(Integer) #: the minimum required key length
    key_lifespan        = Column(Integer) #: the maximal key lifespan in days
    algorithms_str      = Column(String) 

    dead_man_switch     = Column(Boolean)
    description         = Column(UnicodeText)

    def __init__(self, name, implementation_date, ca, key_len, key_lifespan, algorithms, description=None):
        from batzenca.session import session

        self.name = name
        self.implementation_date = implementation_date
        self.ca = ca
        self.key_len = key_len
        self.key_lifespan = key_lifespan

        algs = []
        for alg in algorithms:
            try:
                algs.append(session.gnupg.alg_to_str[alg])
            except KeyError:
                raise ValueError("Algoritm '%s' is unknown. Supported algorithms are '%s'"%(alg, ", ".session.gnupg.str_to_alg.keys()))
        self.algorithms_str = ",".join(algs)
            
        self.description = unicode(description)

    @classmethod
    def from_key(cls, key):
        """Search the database for the policy matching the key.

        :param batzenca.database.keys.Key key: the key to query the database with
        
        .. note::

           The returned object was queried from the main session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(cls).join(Key).filter(Key.kid == key.kid)

        if res.count() == 0:
            raise EntryNotFound("No peer matching key '%s' in database."%key)
        else:
            if res.count() > 1:
                warnings.warn("More than one release with key '%s' found, picking first one."%key)
            return res.first()

    @property
    def algorithms(self):
        """The set of allowed encryption or signature algorithms."""
        from batzenca.session import session
        return set([session.gnupg.str_to_alg[e] for e in self.algorithms_str.split(",")])

    def check_length(self, key):
        """Return ``True`` if the shortest subkey of ``key`` is shorter than the minimum key length
        mandated by this policy.

        :param batzenca.database.keys.Key key: the key to check

        .. note::

            This function issues a :class:`batzenca.database.policies.PolicyViolation` warning if
            ``key``'s length is too short.
        """
        if len(key) < self.key_len:
            msg = u"Key '%s' has key length %d but at least %d is required by '%s'."%(unicode(key), len(key), self.key_len, unicode(self))
            warnings.warn(msg, PolicyViolation)
            return False
        return True

    def check_algorithms(self, key):
        """Return ``True`` if all algorithms used by ``key`` are in the set of allowed algorithms for this policy.

        :param batzenca.database.keys.Key key: the key to check

        .. note::

            This function issues a :class:`batzenca.database.policies.PolicyViolation` warning if
            ``key``'s length is too short.
        """
        from batzenca.session import session

        key_algorithms = set(key.algorithms)
        algorithms = self.algorithms

        if not algorithms:
            return True

        if not key_algorithms.issubset(algorithms):
            diff = key_algorithms.difference(algorithms)
            diff_str = ",".join(session.gnupg.alg_to_str[e] for e in diff)
            msg = u"Key '%s' uses algorithm(s) '%s' which is/are not in '%s' as mandated by '%s'."%(unicode(key), diff_str, self.algorithms_str, unicode(self))
            warnings.warn(msg, PolicyViolation)
            return False
        return True

    def check_expiration(self, key):
        """Return ``True`` if the provided key expires within the time limit mandated by this  policy..

        :param batzenca.database.keys.Key key: the key to check

        .. note::

            This function issues a :class:`batzenca.database.policies.PolicyViolation` warning if
            ``key``'s length is too short.
        """
        if not key.expires and self.key_lifespan > 0:
            msg = u"Key '%s'does not expire but expiry of %d days is mandated by '%s'."%(unicode(key), self.key_lifespan, unicode(self))
            warnings.warn(msg, PolicyViolation)
            return False
        else:
            max_expiry = datetime.date.today() + datetime.timedelta(days=self.key_lifespan)
            if max_expiry < key.expires:
                msg = u"Key '%s' expires on %s but max allowed expiration date is %s."%(unicode(key), key.expires, max_expiry)
                warnings.warn(msg, PolicyViolation)
                return False
            return True

    def check_ca_signature(self, key):
        """Return ``True`` if the provided key is signed by the CA associated with this policy.

        :param batzenca.database.keys.Key key: the key to check

        .. note::

            This function issues a :class:`batzenca.database.policies.PolicyViolation` warning if
            ``key``'s length is too short.
        """
        if not key.is_signed_by(self.ca):
            msg = u"No UID of Key '%s' has a valid signature of the CA key '%s'"%(unicode(key), unicode(self.ca))
            warnings.warn(msg, PolicyViolation)
            return False
        return True

    def check(self, key, check_ca_signature=True):
        """Check if the provided key is valid according to this policy.

        This includes that it's minimum key length, used algorithms and expiration time satisfy the
        constraints set by this policy. Furthermore, the key also must a valid signature by this
        policy's CA unless ``check_signature`` is set to ``False``.

        :param boolean check_ca_signature: if ``False`` this function does not check if the key is
            signed by the CA's key. This might be useful when the CA decides whether to sign this
            key depending on whether it is valid otherwise.

        """
        ret = True
        ret &= self.check_length(key)
        ret &= self.check_algorithms(key)
        ret &= self.check_expiration(key)
        if check_ca_signature:
            ret &= self.check_ca_signature(key)
        return ret

    def __str__(self):
        return "%s: (%d, %d, (%s))"%(self.name, self.key_len, self.key_lifespan, self.algorithms_str)



from distutils.core import setup

setup(
    name = 'BatzenCA',
    version = '0.1',
    packages = ['batzenca', 'batzenca/database'],
    scripts = ['scripts/batzenca-interactive.py', 'scripts/batzenca-release.py'],
    license = 'Simplified BSD License',
    long_description=open('README.rst').read(),
    author = 'Martin R. Albrecht',
    author_email = 'martinralbrech+batzenca@googlemail.com',
    url = 'https://bitbucket.org/malb/batzenca',
    install_requires = open("requirements.txt").read().split("\n"),
)


"""
.. module:: peers
 
.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Peers are people, typically.
"""
from base import Base, EntryNotFound
from keys import Key

from sqlalchemy import Column, Integer, String, Date, Boolean, ForeignKey, UnicodeText, BigInteger
from sqlalchemy.orm import relationship, backref

import warnings

class Peer(Base):
    """This class represents a peer which participates in mailing lists. Typically, a peer
    represents a person which has one or many PGP keys associated with her.

    :param str name: the name of the peer
    :param keys: an instance of :class:`batzenca.database.keys.Key` or a list of such instances
    :param str data0: arbitrary data
    :param str data1: arbitrary data
    :param str data2: arbitrary data
    :param str data3: arbitrary data

    .. note::

       The returned object was not added to any session.
    """
    __tablename__ = 'peers'

    id    = Column(Integer, primary_key=True) #: database id
    name  = Column(String, nullable=False) #: the peer's name
    keys  = relationship('Key', backref=backref('peer'), order_by=Key.timestamp) #: a list of all keys associated with this peer

    data0 = Column(String) #: free form data associated with this peer
    data1 = Column(String) #: free form data associated with this peer
    data2 = Column(String) #: free form data associated with this peer
    data3 = Column(String) #: free form data associated with this peer

    def __init__(self, name, keys, data0='', data1='', data2='', data3=''):
        self.name = name
        if isinstance(keys, Key):
            self.keys = [keys]
        else:
            self.keys = keys

        self.data0 = unicode(data0)
        self.data1 = unicode(data1)
        self.data2 = unicode(data2)
        self.data3 = unicode(data3)

    @classmethod
    def merge(cls, left, right):
        """Merge the two peers ``left`` and ``right`` into a new peer.

        :param batzenca.database.peers.Peer left: first peer
        :param batzenca.database.peers.Peer right: second peer
        
        We favour ``left`` over ``right``. That is, if ``name``, ``email`` or ``dataX`` is set for
        ``left``, we pick this data even if these fields are set in ``right`` as well.
        
        .. note::

           The returned object was not added to any session.
        """
        name = left.name
        keys = set(left.keys).union(right.keys)
        data0 = left.data0 if left.data0 else right.data0
        data1 = left.data1 if left.data1 else right.data1
        data2 = left.data2 if left.data2 else right.data2
        data3 = left.data3 if left.data3 else right.data3

        # calling session.delete and session.add doesn't seem right, call merge_peers function for
        # this.

        return Peer(name, list(keys), data0, data1, data2, data3)

    @classmethod
    def from_key(cls, key):
        """Return the peer associatied with ``key`` in the database.

        :param batzenca.database.keys.Key key: the key to query for
        
        :raises batzenca.database.base.EntryNotFound: when no entry is found
        :raises RuntimeError: if more than one element is found as this is considered an
            inconsistent state of the database

        .. note::

           The returned object was aquired from the master session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(cls).join(Key).filter(Key.kid == key.kid)

        if res.count() == 0:
            raise EntryNotFound("No peer matching key '%s' in database."%key)
        else:
            if res.count() > 1:
                raise RuntimeError("More than one peer associated with key '%s'."%key)
            return res.first()

    @classmethod
    def from_name(cls, name):
        """Return a peer with the given ``name`` from the database.  If more than one element is
        found the "first" element is returned, where "first" has no particular meaning and is
        implementation specific.  In this case a warning is issued.

        :param str name: the name we are looking for

        :raises batzenca.database.base.EntryNotFound: when no entry is found

        .. note::

           The returned object was aquired from the master session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(cls).filter(cls.name == name)

        if res.count() == 0:
            raise EntryNotFound("No peer with name '%s' in database."%name)
        else:
            if res.count() > 1:
                warnings.warn("More than one peer with name '%s' found, picking first one."%name)
            return res.first()

    @classmethod
    def from_email(cls, email):
        """Return a peer with the given ``email`` address from the database. A peer is defined to
        have an e-mail address associated with it, if any of the keys associated with it are for
        said e-mail address.

        If more than one element is found the "first" element is returned, where "first" has no
        particular meaning and is implementation specific.  In this case a warning is issued.

        :param str email: the email address we are looking for

        :raises batzenca.database.base.EntryNotFound: when no entry is found
        
        .. note::

           The returned object was aquired from the master session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(Peer).join(Key).filter(Key.peer_id == Peer.id, Key.email == email)

        if res.count() == 0:
            raise EntryNotFound("No peer with email '%s' in database."%email)
        else:
            if res.count() > 1 and len(res.all()) > 1: # we might find two keys but only one peer
                warnings.warn("More than one peer with email '%s' found, picking first one."%email)
            return res.first()

    def __repr__(self):
        return unicode(u"<Peer: %s, %s>"%(self.id, self.name)).encode('utf-8')

    def __str__(self):
        return unicode(self).encode('utf-8')
        
    def __unicode__(self):
        data = (self.data0, self.data1, self.data2, self.data3)
        data = u", ".join(d for d in data if d)
        if data:
            return u"%s - %s"%(self.name, data)
        else:
            return u"%s"%self.name

    def __hash__(self):
        return hash((self.name, self.data0, self.data1, self.data2, self.data3))

    @property
    def key(self):
        """Return the most recent active key associated with this peer."""
        keys = sorted(self.keys, reverse=True)
        for key in keys:
            if key:
                return key
        return keys[0]

    @property
    def email(self):
        """ Return the e-mail address associated to the most rcent active key associated with this peer."""
        return str(self.key.email)
        
def merge_peers(left, right):
    """Wrapper around :func:`batzenca.database.peers.Peer.merge` which modifies the master session.

    :func:`batzenca.database.peers.Peer.merge` does not modify the master session in any way, but
    this function deletes both ``left`` and ``right`` and adds the result of
    :func:`batzenca.database.peers.Peer.merge`. This result is also returned.

    :param batzenca.database.peers.Peer left: the first peer
    :param batzenca.database.peers.Peer right: the second peer
    
    .. note ::

         Changes to the master session are not committed.

    """
    from batzenca.session import session
    if isinstance(left, int):
        left = session.db_session.query(Peer).filter(Peer.id == left)

    if isinstance(right, int):
        right = session.db_session.query(Peer).filter(Peer.id == right)

    merged = Peer.merge(left, right)
    session.db_session.add(merged)
    session.db_session.delete(left)
    session.db_session.delete(right)
    return merged



"""
.. module:: base
 
.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>
"""

import datetime

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

class EntryNotFound(ValueError):
    """
    This exception is raised if some query returned an empty result.
    """
    pass



"""
.. module:: keys
 
.. moduleauthor:: Martin R. Albrecht <martinralbrecht+batzenca@googlemail.com>

Keys represent PGP keys and are typically stored in the database.
"""
from base import Base, EntryNotFound
from sqlalchemy import Column, Integer, String, Date, ForeignKey
from sqlalchemy.ext.associationproxy import association_proxy

import datetime
import warnings

class Key(Base):
    """This class represents a PGP key. Distributing these keys for mailing lists is the purpose of
    this software package. It is also a design goal of this software package that the
    particularities of dealing with the GnuPG backend are hidden behind the functions of this
    class.

    :param keyid: a key id as an integer or a hex string
    :param str name: a username, if ``None`` is given, then it is read from GnuPG
    :param str email: an e-mail address, if ``None`` is given, then it is read from GnuPG
    :param timestamp: the timestamp when this key was created, if ``None`` is given, then it is
      read from GnuPG.

    .. note::

        The returned object was not added to any session.
    """
    __tablename__ = 'keys'

    id        = Column(Integer, primary_key=True)           #: database id
    kid       = Column(String, nullable=False, unique=True) #: 8 byte pgp key id of the form 0x0123456789abcdef
    name      = Column(String)                              #: user id
    email     = Column(String)                              #: email as stored in key
    timestamp = Column(Date)                                #: date it was added to the database
    peer_id   = Column(Integer, ForeignKey("peers.id"))

    releases    = association_proxy('release_associations', 'release') #: a list of releases this key is in

    @staticmethod
    def canonical_keyid(keyid):
        """Convert keyid to canonical representation as a hex string of the last
        64 bits.

        :param keyid: a key id as an integer or a hex string.

        """
        try:
            keyid = "0x%016x"%keyid
        except TypeError:
            pass
        return "0x%016x"%(int(keyid,16) % (1<<64))
    
    def __init__(self, keyid, name=None, email=None, timestamp=None):
        self.kid = Key.canonical_keyid(keyid)

        from batzenca.session import session

        try:
            _ = session.gnupg.key_get(self.kid)
        except KeyError:
            if name is None or email is None or timestamp is None:
                raise ValueError("The key %s does not exist in GnuPG and not enough information was provided for generating Key instance without it."%self.kid)
            self.name      = unicode(name.strip())
            self.email     = str(email.strip())
            self.timestamp = timestamp
            return

        uid = session.gnupg.key_uid(self.kid)

        self.name = uid.name
        self.email = str(uid.email)
        self.timestamp = session.gnupg.key_timestamp(self.kid)

        try:
            Key.from_keyid(self.kid)
            raise ValueError("Key with keyid '%s' already found in database."%self.kid)
        except EntryNotFound:
            pass
            

    @classmethod
    def from_keyid(cls, keyid):
        """Return the key with ``keyid`` from the database. If no such element is found an
        :class:`batzenca.database.base.EntryNotFound` exception is raised. If more than one element is found this is
        considered an inconsistent state of the database and a :class:`RuntimeError` exception is
        raised.

        :param keyid: key id as integer or a hex string

        .. note::

           The returned object was aquired from the master session and lives there.

        """
        keyid = Key.canonical_keyid(keyid)
            
        from batzenca.session import session
        res = session.db_session.query(cls).filter(cls.kid == keyid)

        if res.count() == 0:
            raise EntryNotFound("No key with key id '%s' in database."%keyid)
        else:
            if res.count() > 1:
                raise RuntimeError("More than one key with key id '%s' in database."%keyid)
            return res.first()

    @classmethod
    def from_name(cls, name):
        """Return a key with the given ``name`` from the database. If no such element is found an
        :class:`batzenca.database.base.EntryNotFound` exception is raised. If more than one element
        is found the "first" element is returned, where "first" has no particular meaning. In this
        case a warning is issued. In particular, no guarantee is given that two consecutive runs
        will yield the same result if more than one key has the provided ``name``.

        :param str name: the name the database is queried for
        
        .. note::

           The returned object was aquired from the master session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(cls).filter(cls.name == name)

        if res.count() == 0:
            raise ValueError("No key with name '%s' in database."%name)
        else:
            if res.count() > 1:
                warnings.warn("More than one key with name '%s' found, picking first one."%name)
            return res.first()

    @classmethod
    def from_email(cls, email):
        """Return a key with the given ``email`` address from the database.

        If no such key is found an :class:`batzenca.database.base.EntryNotFound` exception is
        raised. If more than one element is found the "first" element is returned, where "first" has
        no particular meaning. In this case a warning is issued. In particular, no guarantee is
        given that two consecutive runs will yield the same result if more than one key has the
        provided ``email`` address.

        You can query the database for the most recent key matching an e-mail address you can call::

            Peer.from_email(email).key

        :param str email: the email the database is queried for
        
        .. note::

           The returned object was aquired from the master session and lives there.

        """
        from batzenca.session import session
        res = session.db_session.query(cls).filter(cls.email == email)

        if res.count() == 0:
            raise EntryNotFound("No key with email '%s' in database."%email)
        else:
            if res.count() > 1:
                warnings.warn("More than one key with email '%s' found, picking first one."%email)
            return res.first()

    @classmethod
    def from_filename(cls, filename):
        """Read the file ``filename`` into GnuPG and construct instances of
        :class:`batzenca.database.keys.Key` for each key contained in the file ``filename``. A tuple
        of :class:`batzenca.database.keys.Key` objects is returned.

        :param str filename: a file name
        
        .. note::

           The returned objects were not added to any session, any keys found in ``filename`` were
           added to the GnuPG database.

        """
        from batzenca.session import session

        with open(filename) as fh:
            data = fh.read()
            res = session.gnupg.keys_import(data)
            if len(res) == 0:
                raise EntryNotFound("No key found in in file '%s'"%filename)
            else:
                ret = []
                for fpr in res.keys():
                    keyid = Key.canonical_keyid(int("0x"+fpr[-16:],16))
                    try:
                        key = Key.from_keyid(keyid)
                    except EntryNotFound:
                        key = Key(keyid)
                    ret.append(key)
                return tuple(ret)

    @classmethod
    def from_str(cls, ascii_data):
        """Read the PGP keys in ``ascii_data`` into GnuPG and construct instances of
        :class:`batzenca.database.keys.Key` for each key contained in ``ascii_data``. A tuple of
        :class:`batzenca.database.keys.Key` objects is returned.

        :param str ascii_data: PGP keys in ASCII format
        
        .. note::
        
           The returned object was not added to any session, any keys found in ``ascii_data`` were
           added to the GnuPG database.

        """
        from batzenca.session import session
        res = session.gnupg.keys_import(ascii_data)
        if len(res) == 0:
            cut = "\n".join(ascii_data.splitlines()[:20])
            raise EntryNotFound("""No key found in in provided string\n\n%s"""%cut)
        else:
            ret = []
            for fpr in res.keys():
                keyid = Key.canonical_keyid( int("0x"+fpr[-16:],16) )
                try:
                    key = Key.from_keyid(keyid)
                except EntryNotFound:
                    key = Key(keyid)
                ret.append(key)
            return tuple(ret)

        
    def __nonzero__(self):
        """Return ``True`` if this key has at least one valid (not revoked, expired or disabled)
        subkey for signing and one valid subkey for encrypting.

        """
        from batzenca.session import session
        return session.gnupg.key_okay(self.kid)

    def is_expired(self):
        """Return ``True`` if all subkeys of this key which can be used for encryption are
        expired.

        """
        from batzenca.session import session
        return session.gnupg.key_expired(self.kid)

    def __str__(self):
        return u"%s: %s <%s>"%(self.kid,self.name,self.email)

    def __repr__(self):
        return "<Key: %s %s>"%(self.kid, self.email)

    def __len__(self):
        from batzenca.session import session
        return session.gnupg.key_min_len(self.kid)

    @property
    def expires(self):
        """Expirey date."""
        from batzenca.session import session
        return session.gnupg.key_expires(self.kid)

    @property
    def fingerprint(self):
        """This key's fingerprint as a string in groups of four."""
        from batzenca.session import session
        s = session.gnupg.key_fingerprint(self.kid)
        return " ".join(s[i:i+4] for i in range(0,len(s),4))

    def __lt__(self, other):
        """A key is smaller than another key if it is older."""
        return self.timestamp < other.timestamp

    def __hash__(self):
        return int(self.kid, 16)

    @property
    def algorithms(self):
        from batzenca.session import session
        return session.gnupg.key_pubkey_algos(self.kid)

    def is_signed_by(self, signer):
        from batzenca.session import session
        return session.gnupg.key_any_uid_is_signed_by(self.kid, signer.kid)

    def sign(self, signer):
        from batzenca.session import session
        session.gnupg.key_sign(self.kid, signer.kid)

    def revoke_signature(self, signer, reason=""):
        from batzenca.session import session
        session.gnupg.key_revsig(self.kid, signer.kid, 4, msg=reason)

    def is_valid(self):
        from batzenca.session import session
        return bool(self) and session.gnupg.key_validity(self.kid) > 0

    def hard_delete(self):
        raise NotImplementedError

    def _edit(self):
        """
        .. warning:

           This will open an interactive session using rawinput
        """
        from batzenca.session import session
        return session.gnupg.key_edit(self.kid)

    def delete_signature(self, signer):
        """Delete all signatures made by ``signer``.

        :param batzenca.database.keys.Key signer: - the signer for which to delete signatures.
        
        .. warning:

           This may open an interactive session using rawinput

        """
        from batzenca.session import session
        try:
            return session.gnupg.key_delete_signature(self.kid, signer.kid)
        except AttributeError:
            return session.gnupg.key_delete_signature(self.kid, signer)

    @property
    def signatures(self):
        """A tuple where each entry represents a key that signed this key. The
        entries are either :class:`batzenca.database.keys.Key` objects - if the
        key is present in the local GnuPG key store - or strings of the key
        id.

        """
        from batzenca.session import session
        keyids = tuple(Key.canonical_keyid( keyid ) for keyid in session.gnupg.key_signatures(self.kid))
        sigs = []
        for keyid in keyids:
            if int(self.kid, 16) == int(keyid,16):
                sigs.append(self)
                continue
            try:
                sigs.append(Key.from_keyid(int(keyid,16)))
            except EntryNotFound:
                sigs.append(keyid)
        return tuple(sigs)

    def clean(self, whitelist=None):
        """Remove all signatures except self-signature and those given in
        whitelist.

        :param iterable whitelist: a list of
            :class:`batzenca.database.keys.Key`,
            :class:`batzenca.database.peers.Peer`,
            :class:`batzenca.database.policies.Policy`,
            :class:`batzenca.database.mailinglists.MailingList`.
            :class:`batzenca.database.releases.Release`.
        """
        from keys import Key
        from peers import Peer
        from mailinglists import MailingList
        from policies import Policy
        from releases import Release

        def extract_key(obj):
            if isinstance(obj, str) or isinstance(obj, unicode):
                return obj
            elif isinstance(obj, Key):
                return obj
            elif isinstance(obj, Peer):
                return obj.Key
            elif isinstance(obj, Policy):
                return obj.ca
            elif isinstance(obj, MailingList):
                return obj.policy.ca
            elif isinstance(obj, Release):
                return obj.policy.ca
            else:
                raise TypeError("Type '%s' of '%s' not understood."%(type(obj),obj))

        if whitelist:
            whitelist = [extract_key(obj) for obj in whitelist]
        else:
            whitelist = []
        whitelist.append(self)

        for signature in set(self.signatures).difference(whitelist):
            self.delete_signature(signature)
        
    @property
    def _pyme_key(self):
        from batzenca.session import session
        return session.gnupg.key_get(self.kid)